package org.openxma.xmadsl;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.openarchitectureware.util.stdlib.ElementPropertiesExtensions;
import org.openxma.xmadsl.model.Attribute;
import org.openxma.xmadsl.model.AttributeHolder;
import org.openxma.xmadsl.model.AttributeWithProperties;
import org.openxma.xmadsl.model.ComplexType;
import org.openxma.xmadsl.model.CustomizeableField;
import org.openxma.xmadsl.model.DataBaseConstraint;
import org.openxma.xmadsl.model.DataBaseField;
import org.openxma.xmadsl.model.DataTypeAndTypeParameter;
import org.openxma.xmadsl.model.DataView;
import org.openxma.xmadsl.model.DataViewSource;
import org.openxma.xmadsl.model.Entity;
import org.openxma.xmadsl.model.IField;
import org.openxma.xmadsl.model.IncludeExcludeDefinition;
import org.openxma.xmadsl.model.Reference;
import org.openxma.xmadsl.model.ValueObject;

public class XmadslModelExtension {
	
	private static final int ALL = 1;
	private static final int NATURAL = 2;
	private static final int ID = 3;
	private static final int REFERENCE = 4;		
	private static final int ATTRIBUTE = 5;		
	
	/**
	 * Get the label text for a CustomizeableField
	 * @param field
	 * @return
	 */
	public static String getLabelText(CustomizeableField field) {				
		if (field == null) throw new IllegalArgumentException();
		//System.out.println(">>>> getLabelText for field: "+field.getObject().getName()+"."+field.getAttributeHolder().getName());
		return field.getLabelText();
	}
	
	/**
	 * Get the label text for a CustomizeableField
	 * @param field
	 * @return
	 */
	public static String getTooltipText(CustomizeableField field) {				
		if (field == null) throw new IllegalArgumentException();
		return field.getTooltipText();
	}		

	/**
	 * Get the unit text for a CustomizeableField
	 * @param field
	 * @return
	 */
	public static String getUnitText(CustomizeableField field) {		
		if (field == null) throw new IllegalArgumentException();
		return field.getUnitText();
	}
	/**
	 * Get the unit-attribute for a CustomizeableField
	 * @param field
	 * @return
	 */
	public static Attribute getUnitAttribute(CustomizeableField field) {		
		if (field == null) throw new IllegalArgumentException();
		return field.getUnitAttribute();
	}
	
	/**
	 * Get the unit-object for a CustomizableField
	 * @param field
	 * @return
	 */
	public static Object getUnitObject(CustomizeableField field) {
		if (field == null) throw new IllegalArgumentException();
		return field.getUnitObject();
	}
	
	public static String getName(AttributeHolder attributeHolder) {
		if (attributeHolder != null) {		
			return attributeHolder.getName();
		}
		return null;
	}
	
	public static String getDoc(AttributeHolder attributeHolder) {
		if (attributeHolder != null) {		
			return attributeHolder.getAttribute().getDoc();
		}
		return null;
	}
	
	public static DataTypeAndTypeParameter getType(AttributeHolder attributeHolder) {
		if (attributeHolder != null) {		
			return attributeHolder.getAttribute().getType();
		}
		return null;
	}		
	
	public static String getAttributeName(IField iField) {
		if (iField != null && iField.getAttributeHolder() != null) {
			return iField.getAttributeHolder().getName();
		}
		return null;
	}
	
	public static Attribute getAttribute(IField iField) {
		if (iField != null && iField.getAttributeHolder() != null) {
			return iField.getAttributeHolder().getAttribute();
		}
		return null;
	}
	
	/**
	 * Get the Attribute hold by an AttributeHolder
	 * @param attrHolder
	 * @return
	 */
	public static Attribute getAttributeOfAttributeHolder(AttributeHolder attrHolder) {
		
		if (attrHolder instanceof Attribute) {
			return (Attribute)attrHolder;
		}
		if (attrHolder instanceof AttributeWithProperties) {
			AttributeHolder nestedAttributeHolder = ((AttributeWithProperties)attrHolder).getAttributeHolder();
			if (nestedAttributeHolder != null) {
				return getAttributeOfAttributeHolder(nestedAttributeHolder);
			} else {
				throw new RuntimeException("Nested AttributeHolder is null");
			}
		}
		throw new RuntimeException("Type of AttributeHolder not supported: "+attrHolder.getClass());
	}
	
	/**
	 * 
	 * @param data
	 * @return
	 */
	public static List<AttributeHolder> allAttributeHolders(ComplexType data) {
		
		List<AttributeHolder> attrList = null;
		
//		Object prop = ElementPropertiesExtensions.getProperty(data,"attrs");		
//		if (prop != null && prop instanceof List) {
//			attrList = (List<AttributeHolder>)prop;
//			return attrList;			
//		}
		if (data instanceof DataView) {
			attrList=allAttributeHoldersDataView((DataView)data);		
		} else if (data instanceof Entity) {
			attrList= allAttributeHoldersEntity((Entity)data);
		} else if (data instanceof ValueObject) {
			attrList= allAttributeHoldersValueObject((ValueObject)data);								
		} else {
			throw new RuntimeException("Default implementation of allAttributes used for: "+data);
		}
//		ElementPropertiesExtensions.setProperty(data,"attrs",attrList);		
		return attrList;
		
	}
    /**
	 * Create list of all attributes of a DataObject (including the attributes
	 * of the "base" DataObject)
	 */	
	public static List<AttributeHolder> allAttributeHoldersEntity(Entity data) {
		List<AttributeHolder> attrList = new ArrayList<AttributeHolder>();
		for (Attribute attr: data.getAttributes()) {
			attrList.add(attr);
		}
		if (data.getSuperType() != null) {
			List<AttributeHolder> attrOfSuperType = allAttributeHolders(data.getSuperType()); 					
			attrList.addAll(attrOfSuperType);
		}
		return attrList;
	}	
			
	/** Create list of all attributes of a ValueObject
	 * 
	 * @param valueObject
	 * @return
	 */
	public static List<AttributeHolder> allAttributeHoldersValueObject(ValueObject valueObject) {
		List<AttributeHolder> attrList = new ArrayList<AttributeHolder>();
		for (Attribute attr: valueObject.getAttributes()) {
			attrList.add(attr);
		}
		return 	attrList;
	}
	
	/**
	 * 
	 * @param data
	 * @return
	 */
	public static List<AttributeHolder> allAttributeHoldersDataView(DataView data) {
		
		List<AttributeHolder> attrList = new ArrayList<AttributeHolder>();
		EList<IncludeExcludeDefinition> includes = data.getIncludes();
		EList<DataViewSource> sources = data.getSourceDefinition().getSources();
		
		for (IncludeExcludeDefinition incExcDef : includes) {			
			List<AttributeHolder> tempList = allAttributeHolders(incExcDef,sources);
			if (tempList != null) {
				attrList.addAll(tempList);
			}
		}
		return attrList;
	}

	/**
	 * 
	 * @param include
	 * @param sources
	 * @return
	 */
	public static List<AttributeHolder> allAttributeHolders(IncludeExcludeDefinition include, EList<DataViewSource> sources) {		
		int incExcType = getIncExcType(include);
		ComplexType firstSource = null;
		if (sources != null && sources.size()>0) {
			firstSource = ((DataViewSource)sources.get(0)).getSource();
		}
		if (firstSource == null) {
			return null;
		}
		switch(incExcType) {
			case ALL: 
				return allAttributeHolders(firstSource);
			case NATURAL: 
				if (firstSource instanceof Entity == false) {
					throw new RuntimeException("Natural attribute include only supported for Entity");
				}
				Entity entity = (Entity)firstSource;
				DataBaseConstraint dbConstraint = getNaturalIdentifier(entity);
				return getResolvedAttributeList(dbConstraint);								
			case ID: 
				return getAttributeListHolderForName(allAttributeHolders(firstSource),"oid");
			case REFERENCE: 
				return new ArrayList<AttributeHolder>();		
			case ATTRIBUTE:
				List<AttributeHolder> tempAttrList = new ArrayList<AttributeHolder>();				
				tempAttrList.add(include.getAttributeWithProperties());
				return tempAttrList;
			default: throw new RuntimeException("Unsupported type of IncludeExclude for "+((ComplexType)include.eContainer()).getName());
		}
	}
	
	/**
	 * 
	 * @param attributeHolders
	 * @param name
	 * @return
	 */
	private static List<AttributeHolder> getAttributeListHolderForName(List<AttributeHolder> attributeHolders, String name) {
		for (AttributeHolder attrHolder: attributeHolders) {
			if (attrHolder.getName().equals(name)) {
				List<AttributeHolder> tempList = new ArrayList<AttributeHolder>();
				tempList.add(attrHolder);						
				return tempList;
			}
		}
		return null;
	}

	/**
	 * 
	 * @param incExcDef
	 * @return
	 */
	private static int getIncExcType(IncludeExcludeDefinition incExcDef) {
		
		if (incExcDef.isAll()) return ALL;
		if (incExcDef.isNatural()) return NATURAL;
		if (incExcDef.isId()) return ID;		
		if (incExcDef.getReference() != null) return REFERENCE;
		return ATTRIBUTE;
	}
	
	/**
	 *  Returns the natural identifier
	 */	
	public static DataBaseConstraint getNaturalIdentifier(Entity entity) {
	 	Entity baseEntity = (entity.getSuperType() == null  ? entity : entity.getSuperType());
	 	EList<DataBaseConstraint> dbConstraints = baseEntity.getPersistence().getDataBaseConstraints();
	 	for (DataBaseConstraint constr : dbConstraints) {
	 		if (constr.getType().equals("primary-key")) {
	 			return constr;
	 		}
	 	}
	 	return null;
	}
	
	/**
	 *  Resolves all natural identifier attributes of the given dataBaseConstraint
	 */	
	public static List<AttributeHolder> getResolvedAttributeList(DataBaseConstraint dataBaseConstraint) {
		List<AttributeHolder> allAttributes = new ArrayList<AttributeHolder>();
		for (DataBaseField dbField: dataBaseConstraint.getFields()) {
			List<AttributeHolder> tempList = getResolvedAttributeList(dbField);
			if (tempList != null) {
				allAttributes.addAll(tempList);
			}
		}		
		for (AttributeHolder attribute : allAttributes) {
			for (AttributeHolder e : allAttributes) {
				if (e.getName().equals(attribute.getName()) && e != attribute && getEntity(e) != getEntity(attribute)) {
					setFQNProperty(attribute.getAttribute(),getEntity(attribute).getName()); 
				} 
			}
		}
		return allAttributes;
	}
	
	/**
	 * 
	 * @param entity
	 * @return
	 */
	public static List<AttributeHolder> getResolvedAttributeList(Entity entity) {
		EList<DataBaseField> fields = getNaturalIdentifier(entity).getFields();
		List<AttributeHolder> attrList = new ArrayList<AttributeHolder>();			
		if (fields != null) {					
			for (DataBaseField field: fields) {
				attrList.addAll(getResolvedAttributeList(field));	
			}
		}
		return attrList;
	}
	
	/**
	 * DataBaseField: entity.getNaturalIdentifier().fields.getResolvedAttributeList()
	 * Reference: ((Entity)reference.type).getNaturalIdentifier().getResolvedAttributeList()
	 * @param field
	 * @return
	 */
	public static List<AttributeHolder> getResolvedAttributeList(DataBaseField field) {
		
		List<AttributeHolder> attrList = new ArrayList<AttributeHolder>();
		
		if (field instanceof Attribute) {
			Attribute attr = (Attribute)field;			
			attrList.add(attr);			
			return attrList;
		}						
		if (field instanceof Reference) {
			Reference ref = (Reference)field;
			Entity entity = (Entity)ref.getType();
			DataBaseConstraint dbConstraint = getNaturalIdentifier(entity);
			return getResolvedAttributeList(dbConstraint);			
		}
		return attrList;
	}
	
	/**
	 * 
	 * @param attributeHolder
	 * @return
	 */
	public static Entity getEntity(AttributeHolder attributeHolder) {
		Attribute attr = attributeHolder.getAttribute();
		EObject container = attr.eContainer();
		if (container instanceof Entity) {
			return (Entity)container;
		}
		return null;
	}
	
	public static void setFQNProperty(AttributeHolder attr, String entityName) {		
		ElementPropertiesExtensions.setProperty(attr,"fqn",entityName);
	}
		
	public static String getFQNProperty(AttributeHolder attr) {		
		return (String)ElementPropertiesExtensions.getProperty(attr,"fqn");
	}
	
	public static void setFQNProperty(Reference ref, String entityName) {		
		ElementPropertiesExtensions.setProperty(ref,"fqn",entityName);
	}
		
	public static String getFQNProperty(Reference ref) {		
		return (String)ElementPropertiesExtensions.getProperty(ref,"fqn");
	}	
	
	

	
}
