/*******************************************************************************
 * Copyright (c) 2008 itemis AG (http://www.itemis.eu) and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/
package org.openxma.xmadsl.linking.impl;

import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.antlr.runtime.Token;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.openarchitectureware.xtext.parser.parsetree.Node;
import org.openxma.xmadsl.scoping.IScope;
import org.openxma.xmadsl.scoping.IScopeProvider;
import org.openxma.xmadsl.scoping.IScopedElement;
import org.openxma.xmadsl.scoping.impl.XmaDslScopeProvider;

/**
 * @author Heiko Behrens - Initial contribution and API
 * @author Michael Clay
 * @author Sebastian Zarnekow
 */
public class DefaultLinkingService extends AbstractLinkingService {

	private IScopeProvider scopeProvider = new XmaDslScopeProvider();


	protected IScope getScope(EObject context, EReference reference) {
		if (scopeProvider == null)
			throw new IllegalStateException("scopeProvider must not be null.");
		return scopeProvider.getScope(context, reference);
	}

	/**
	 * @return the first element returned from the injected
	 *         {@link IScopeProvider} which matches the text of the passed
	 *         {@link LeafNode}
	 */
	public List<EObject> getLinkedObjects(EObject context, EReference ref,
			Node node) throws IllegalNodeException {
		final EClass requiredType = ref.getEReferenceType();
		if (requiredType == null)
			return Collections.<EObject> emptyList();

		final IScope scope = getScope(context, ref);
		final Iterator<IScopedElement> iterator = scope.getAllContents()
				.iterator();
		final String s = getCrossRefNodeAsString(node);
		if (s != null) {
			while (iterator.hasNext()) {
				final IScopedElement element = iterator.next();
				if (s.equals(element.name()))
					return Collections.singletonList(element.element());
			}
		}
		return Collections.emptyList();
	}

	public String getCrossRefNodeAsString(Node node)
			throws IllegalNodeException {
		String convertMe = null;
		if (node.getChildren().isEmpty())
			convertMe = ((Node) node).getToken().getText();
		else {
			StringBuilder builder = new StringBuilder();
			boolean hiddenSeen = false;
			EList children = node.getChildren();
			for (Iterator iterator = children.iterator(); iterator.hasNext();) {
				Node leaf = (Node) iterator.next();
				if (leaf.getToken().getChannel() != Token.HIDDEN_CHANNEL) {
					if (hiddenSeen && builder.length() > 0)
						builder.append(' ');
					builder.append(leaf.getToken().getText());
					hiddenSeen = false;
				} else {
					hiddenSeen = true;
				}
			}
			convertMe = builder.toString();
		}
		return convertMe;
	}

	/**
	 * @return the name of the first {@link IScopedElement} returned from the
	 *         injected {@link IScopeProvider} for the passed object
	 *         {@link EObject}
	 */
	public String getLinkText(EObject object, EReference reference,
			EObject context) {
		return getUnconvertedLinkText(object, reference, context);
	}

	protected String getUnconvertedLinkText(EObject object,
			EReference reference, EObject context) {
		IScope scope = scopeProvider.getScope(context, reference);
		if (scope == null)
			return null;
		IScopedElement scopedElement = getScopedElement(scope, object);
		if (scopedElement == null)
			return null;
		return scopedElement.name();
	}

	private IScopedElement getScopedElement(IScope scope, EObject element) {
		Iterable<IScopedElement> allContents = scope.getAllContents();
		URI left = EcoreUtil.getURI(element);
		for (IScopedElement scopedElement : allContents) {
			URI right = EcoreUtil.getURI(scopedElement.element());
			if (left.equals(right))
				return scopedElement;
		}
		return null;
	}

	public void setScopeProvider(IScopeProvider scopeProvider) {
		this.scopeProvider = scopeProvider;
	}

	public IScopeProvider getScopeProvider() {
		return scopeProvider;
	}

}
