/*******************************************************************************
 * Copyright (c) 2008 itemis AG (http://www.itemis.eu) and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 *******************************************************************************/
package org.openxma.xmadsl.linking.lazy;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.common.util.WrappedException;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.openarchitectureware.xtext.parser.parsetree.Node;
import org.openxma.xmadsl.linking.ILinkingService;
import org.openxma.xmadsl.linking.impl.DefaultLinkingService;
import org.openxma.xmadsl.linking.impl.XtextLinkingDiagnostic;
import org.openxma.xmadsl.resource.xmadslResource;
import org.openxma.xmadsl.util.Triple;

/**
 * @author Sven Efftinge - Initial contribution and API
 */
public class LazyLinkingResource extends xmadslResource {
	

	public static String OPTION_RESOLVE_ALL = xmadslResource.class.getName() + ".RESOLVE_ALL";
	
	private ILinkingService linkingService = new DefaultLinkingService();
	
	private LazyURIEncoder encoder = new LazyURIEncoder();
	
	public LazyLinkingResource(URI uri) {
		super(uri);
	}

	@Override
	protected void doLoad(InputStream inputStream, Map<?, ?> options) throws IOException {
		super.doLoad(inputStream, options);
		if (options != null && Boolean.TRUE.equals(options.get(OPTION_RESOLVE_ALL)))
			EcoreUtil.resolveAll(this);
	}
	
	@Override
	public EObject getEObject(String uriFragment) {
		try {
			if (getEncoder().isCrossLinkFragment(this, uriFragment)) {
				Triple<EObject, EReference, Node> triple = getEncoder().decode(this, uriFragment);
				List<EObject> linkedObjects = getLinkingService().getLinkedObjects(triple.getFirst(), triple.getSecond(),
						triple.getThird());
				if (linkedObjects.isEmpty()) {
					XtextLinkingDiagnostic diag = createDiagnostic(triple);
					if (!getErrors().contains(diag))
						getErrors().add(diag);
					return null;
				}
				if (linkedObjects.size() > 1)
					throw new IllegalStateException("linkingService returned more than one object for fragment "
							+ uriFragment);
				EObject result = linkedObjects.get(0);
				if (!triple.getSecond().getEReferenceType().isSuperTypeOf(result.eClass())) {
					XtextLinkingDiagnostic diag = createDiagnostic(triple);
					if (!getErrors().contains(diag))
						getErrors().add(diag);
					return null;
				}
				// remove previously added error markers, since everything should be fine now
				XtextLinkingDiagnostic diag = createDiagnostic(triple);
				getErrors().remove(diag);
				return result;
			}
		} catch (RuntimeException e) { 
			e.printStackTrace(); 
			// wrapped because the javaDoc of this method states that WrappedExceptions are thrown
			throw new WrappedException(e);
		}
		return super.getEObject(uriFragment);
	}

	private XtextLinkingDiagnostic createDiagnostic(Triple<EObject, EReference, Node> triple) {
		String msg = "Couldn't resolve reference to "+triple.getSecond().getEType().getName()+" "+triple.getThird().toString();
		return new XtextLinkingDiagnostic(triple.getThird(), msg);
	}

	public void setLinkingService(ILinkingService linkingService) {
		this.linkingService = linkingService;
	}

	public ILinkingService getLinkingService() {
		return linkingService;
	}

	public void setEncoder(LazyURIEncoder encoder) {
		this.encoder = encoder;
	}

	public LazyURIEncoder getEncoder() {
		return encoder;
	}
}
