/*******************************************************************************
 * Copyright (c) 2008 itemis AG (http://www.itemis.eu) and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 *******************************************************************************/
package org.openxma.xmadsl.scoping.impl;

import static com.google.common.collect.Iterables.filter;
import static com.google.common.collect.Iterables.transform;

import java.util.Collection;
import java.util.Iterator;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.openarchitectureware.emf.EcoreUtil2;
import org.openxma.xmadsl.scoping.IScopedElement;

import com.google.common.base.Function;
import com.google.common.base.Predicate;
import com.google.common.collect.Iterators;
import com.google.common.collect.ListMultimap;
import com.google.common.collect.Multimaps;

/**
 * @author Sven Efftinge - Initial contribution and API
 * 
 */
public abstract class ScopedElements {
	private ScopedElements() {
	}

	/**
	 * returns an Iterable<IScopedElement> containing all scoped elements for the AllContnts of the passed resource,
	 * which are of the given type.
	 * The name is computed using the passed function
	 * 
	 * 
	 * @param resource
	 * @param typeToFilter
	 * @param nameFunc
	 * @return
	 */
	public static Iterable<IScopedElement> allInResource(final Resource resource, final EClass typeToFilter,
			final Function<EObject, String> nameFunc) {
		return allInResource(resource, new Predicate<EObject>() {
			public boolean apply(EObject input) {
				return typeToFilter.isInstance(input);
			}
		}, nameFunc);
	}
	
	/**
	 * 
	 * @param parent
	 * @param typeToFilter
	 * @param nameFunc
	 * @return
	 */
	public static Iterable<IScopedElement> allInEObject(final EObject parent, final EClass typeToFilter,
			final Function<EObject, String> nameFunc) {
		return allInCollection(EcoreUtil2.allContents(parent), new Predicate<EObject>() {
			public boolean apply(EObject input) {
				return typeToFilter.isInstance(input);
			}
		}, nameFunc);
	}
	
	/**
	 * 
	 * @param parent
	 * @param typeToFilter
	 * @param nameFunc
	 * @return
	 */
	public static Iterable<IScopedElement> allInCollection(Collection<? extends EObject> collection, final EClass typeToFilter,
			final Function<EObject, String> nameFunc) {
		return allInCollection(collection, new Predicate<EObject>() {
			public boolean apply(EObject input) {
				return typeToFilter.isInstance(input);
			}
		}, nameFunc);
	}

	/**
	 * @param resource
	 * @param predicate
	 * @param nameFunc
	 * 
	 * skips duplicate elements
	 * 
	 * @return
	 */
	public static Iterable<IScopedElement> allInResource(final Resource resource, final Predicate<EObject> predicate,
			final Function<EObject, String> nameFunc) {
		return allInResource(resource,predicate,nameFunc, true);
	}
	
	
	/**
	 * 
	 * @param parent
	 * @param predicate
	 * @param nameFunc
	 * @return
	 */
	public static Iterable<IScopedElement> allInCollection(Collection<? extends EObject> collection, final Predicate<EObject> predicate,
			final Function<EObject, String> nameFunc) {
		return allInCollection(collection,predicate,nameFunc, true);
	}
	
	
	/**
	 * @param resource
	 * @param filter
	 * @param nameFunc
	 * @param skipDuplicates
	 * @return
	 */
	public static Iterable<IScopedElement> allInResource(final Resource resource, final Predicate<EObject> filter,final Function<EObject, String> nameFunc, boolean skipDuplicates) {
		if (resource != null) {
			
			Iterable<EObject> iterable = new Iterable<EObject>(){
			
				public Iterator<EObject> iterator() {
					return EcoreUtil.getAllContents(resource, true);
				}
			};
			
			return filterAndTransform(iterable, filter, nameFunc,
					skipDuplicates);
		}
		return new Iterable<IScopedElement>() {
			public Iterator<IScopedElement> iterator() {
				return Iterators.emptyIterator();
			}};
	}
	
	
	private static Iterable<IScopedElement> allInCollection(final Collection<? extends EObject> collection, final Predicate<EObject> filter,final Function<EObject, String> nameFunc, boolean skipDuplicates) {
		if (collection != null) {
			
			Iterable<EObject> iterable = new Iterable<EObject>(){
			
				public Iterator<EObject> iterator() {
					return (Iterator<EObject>) collection.iterator();
				}
			};
			
			return filterAndTransform(iterable, filter, nameFunc,
					skipDuplicates);
		}
		return new Iterable<IScopedElement>() {
			public Iterator<IScopedElement> iterator() {
				return Iterators.emptyIterator();
			}};
	}

	private static Iterable<IScopedElement> filterAndTransform(
			Iterable<EObject> iterable, final Predicate<EObject> filter,
			final Function<EObject, String> nameFunc, boolean skipDuplicates) {
		Iterable<EObject> filtered = filter(iterable, filter);
		
		Iterable<IScopedElement> transformed = transform(filtered, new Function<EObject, IScopedElement>(){

			public IScopedElement apply(EObject from) {
				IScopedElement scopedElement = ScopedElement.create(nameFunc.apply(from), from);
				return scopedElement;
			}
		});
		if (!skipDuplicates) 
			return transformed;
		
		final ListMultimap<String,IScopedElement> multiMap = Multimaps.index(transformed, new Function<IScopedElement, String>(){
			public String apply(IScopedElement from) {
				return from.name();
			}
		});
		
		return filter(transformed, new Predicate<IScopedElement>(){
			public boolean apply(IScopedElement input) {
				return multiMap.get(input.name()).size()==1;
			}
		});
	}

}
