package org.openxma.xmadsl.validation;

import static org.openxma.xmadsl.scoping.impl.XmaDslScopeProvider.NAME_RESOLVER;
import static org.eclipse.emf.ecore.util.EcoreUtil.*;
import static com.google.common.collect.Multimaps.*;
import static com.google.common.collect.Collections2.*;
import static com.google.common.collect.Iterators.*;

import java.util.Collection;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EValidator;
import org.openxma.xmadsl.XmadslModelExtension;
import org.openxma.xmadsl.model.Attribute;
import org.openxma.xmadsl.model.DataType;
import org.openxma.xmadsl.model.IField;
import org.openxma.xmadsl.model.ParameterValue;
import org.openxma.xmadsl.model.ReferencedXMAComposite;
import org.openxma.xmadsl.model.SimpleType;
import org.openxma.xmadsl.model.Validator;
import org.openxma.xmadsl.model.XMAWidgetEventMapping;
import org.openxma.xmadsl.model.XmadslComposite;
import org.openxma.xmadsl.model.XmadslPackage;

import at.spardat.xma.guidesign.XMAButton;
import at.spardat.xma.guidesign.XMATree;
import at.spardat.xma.guidesign.flex.IDefSelectFunctionCaller;
import at.spardat.xma.guidesign.flex.ISelectFunctionCaller;

import com.google.common.base.Function;
import com.google.common.base.Predicate;
import com.google.common.collect.ListMultimap;
import com.google.common.collect.UnmodifiableIterator;


/**
 * Represents {@link EValidator} implementation used instead of the traditional Check files from oAW4.3.1.
 */
public class XmaDslValidator extends AbstractXmaDslValidator {
	@Check(CheckType.NORMAL)
	public void checkIfAttributeIsUnique(final Attribute attribute) {
		Collection<EObject> attributes = getObjectsByType(attribute.eContainer().eContents(), XmadslPackage.eINSTANCE.getAttribute());
		final ListMultimap<String,? extends EObject> multiMap = index(attributes, NAME_FUNCTION);
		Collection<EObject> duplicateAttributeNames = filter(attributes, new Predicate<EObject>(){
			public boolean apply(EObject input) {
				return input.equals(attribute) && multiMap.get(NAME_RESOLVER.apply(input)).size()>1;
			}
		});
		if (!duplicateAttributeNames.isEmpty()) {
			error("Duplicate attribute '"+attribute.getName()+"'", XmadslPackage.IELEMENT_WITH_NAME__NAME);
		}
	}
	
	@Check(CheckType.NORMAL)
	public void checkDataTypeUnique(final DataType dataType){
		final String dataTypeId = NAME_RESOLVER.apply(dataType);
		UnmodifiableIterator<EObject> dataTypes = filter(getRootContainer(dataType).eAllContents(), new Predicate<EObject>(){
			public boolean apply(EObject eObject) {
				return eObject instanceof DataType && !dataType.equals(eObject) && NAME_RESOLVER.apply(eObject).equals(dataTypeId);
			}});
		if (dataTypes.hasNext()) {
			error("Duplicate type '"+dataTypeId+"' detected",null);
		}
	} 
	
	@Check(CheckType.NORMAL)
	public void checkValidatorUnique(final Validator validator){
		final String validatorId = NAME_RESOLVER.apply(validator);
		UnmodifiableIterator<EObject> validators = filter(getRootContainer(validator).eAllContents(), new Predicate<EObject>(){
			public boolean apply(EObject eObject) {
				return eObject instanceof Validator &&  !validator.equals(eObject) && NAME_RESOLVER.apply(eObject).equals(validatorId);
			}});
		if (validators.hasNext()) {
			error("Duplicate validator '"+validatorId+"' detected",null);
		}
	} 
	
	@Check(CheckType.NORMAL)
	public void checkCompositeSimpleLayout(ReferencedXMAComposite composite) {
		boolean check = composite instanceof XmadslComposite ||  ((ReferencedXMAComposite)composite).getXmaComposite().isYnSimpleLayout();
		if (!check) {
			error("Composite '" + NAME_RESOLVER.apply(composite) + "' has no SimpleLayout and can't be modelled",null);
		}
	}
	
	@Check(CheckType.NORMAL)
	public void checkField(IField field) {
		if (!XmadslModelExtension.allAttributeHolders(field.getObject().getType()).contains(field.getAttributeHolder())) {
			error("Attribute '"+NAME_RESOLVER.apply(field.getAttributeHolder())+"' is not a member of '" +NAME_RESOLVER.apply(field.getObject().getType()) + "'",XmadslPackage.IFIELD__ATTRIBUTE_HOLDER);
		}
	}
	
	@Check(CheckType.NORMAL)
	public void checkTypeDef(SimpleType simpleType) {
		Collection<String> validatorParameterNames = transform(simpleType.getValidatorReference().getValidatorParameter(),new Function<ParameterValue, String>() {
			public String apply(ParameterValue parameterValue) {
				return NAME_RESOLVER.apply(parameterValue.getReferencedParameter());
			}
		});
		Collection<String> typeParameterNames = transform(simpleType.getTypeParameter(),NAME_FUNCTION);
		if (!validatorParameterNames.containsAll(typeParameterNames)) {
			error("Parameter reference for validator not defined in type '"+NAME_RESOLVER.apply(simpleType)+"'",XmadslPackage.SIMPLE_TYPE__TYPE_PARAMETER);
		}
	}
	
	@Check(CheckType.NORMAL)
	public void checkControl(XMAWidgetEventMapping widgetEventMapping) {
		if (!(widgetEventMapping.getControl() instanceof ISelectFunctionCaller)
				&& !(widgetEventMapping.getControl() instanceof XMATree)) {
			error("Type of widget '" + NAME_RESOLVER.apply(widgetEventMapping) + "' does not support event mapping!",XmadslPackage.XMA_WIDGET_EVENT_MAPPING__CONTROL);
		}
	};		
	
	@Check(CheckType.NORMAL)
	public void checkEventType(XMAWidgetEventMapping widgetEventMapping) {
		boolean check=false;
		switch (widgetEventMapping.getEvent()) {
		case ON_SELECTION: check = widgetEventMapping.getControl() instanceof ISelectFunctionCaller;
		case ON_DEFAULT_SELECTION: check = widgetEventMapping.getControl() instanceof IDefSelectFunctionCaller;
		case ON_EXPAND: check = widgetEventMapping.getControl() instanceof XMATree;
		case ON_COLLAPSE: check = widgetEventMapping.getControl() instanceof XMATree;
		case NULL: check = widgetEventMapping.getControl() instanceof XMAButton;
		default: check=false;
		}
		if (!check) {
			error("Widget '" + NAME_RESOLVER.apply(widgetEventMapping.getControl()) + "' does not support event '"
					+ widgetEventMapping.getEvent().getName() + "'",
					XmadslPackage.XMA_WIDGET_EVENT_MAPPING__EVENT);
		}
	};		
	
	// name function used to reolve EObject id
	private final static Function<EObject, String> NAME_FUNCTION = new Function<EObject, String>() {
		public String apply(EObject input) {
			return NAME_RESOLVER.apply(input);
		}
	};
	
}
